/*
	ColorLooks plugin for AviSynth -- This is a collection of filters.
	They are based on Trevlac's (ColourLooks) and Graft's (Colorize) VirtualDub plugins.

	Copyright (C) 2005 Wilbert Dijkhof
  
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.
	 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
		
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
	
	The author can be contacted at:
	Wilbert Dijkhof
	wilbertdijkhof at yahoo.com
*/

#include "ColorLooks.h"
#include "convert.h"

Technicolor::Technicolor(PClip _child, const char* _channel, IScriptEnvironment* env) :
	GenericVideoFilter(_child), channel(_channel) {

	if (!vi.IsRGB()) {
		env->ThrowError("Technicolor: your source is not RGB");
	}

	if ((lstrcmpi(channel, "R") == 0) && (lstrcmpi(channel, "G") == 0) && (lstrcmpi(channel, "B") == 0)) {
		env->ThrowError("Technicolor: channel should be \"R\", \"G\" or \"B\"");
	}
}

Technicolor::~Technicolor() {
}

PVideoFrame __stdcall Technicolor::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
		
	const unsigned char* srcp = src->GetReadPtr();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	
	PVideoFrame dst = env->NewVideoFrame(vi);				
	unsigned char* dstp = dst->GetWritePtr();
	int dst_pitch = dst->GetPitch();
	int dst_width = dst->GetRowSize();
	int dst_height = dst->GetHeight();

	// GB: cyan, RB: magenta, RG: yellow
	int r, g, b, c, m, y;
	int a;

	if (vi.IsRGB32()) {
		a=4;
	} else if (vi.IsRGB24()) {
		a=3;
	}

/*	// Techno Lookup
	int techno[256][256];
	// combine two colors, for example blue and green
	// bd = sqrt(b*b + g*g) * (255.0/360.0);
	for (int b2=0; b2<256; b2++) {
		for (int g2=b2; g2<256; g2++) {
			techno[b2][g2] = (int) (sqrt(b2*b2 + g2*g2) * (255.0/360.0));
			techno[g2][b2] = techno[b2][g2];
		}
	}*/

	// Technicolor
	if (lstrcmpi(channel, "R") == 0) {
		for (int h=0; h < src_height; h++) {
			for (int w=0; w < src_width; w+=a) {
				b = *(srcp +w);
				g = *(srcp +w+1);
				r = *(srcp +w+2);
				c = (int) (sqrt(b*b + g*g) * (255.0/360.0)); // cyan
				*(dstp +w) = c;
				*(dstp +w+1) = c;
				*(dstp +w+2) = r;
			}															
			srcp += src_pitch;
			dstp += dst_pitch;
		}
	} else if (lstrcmpi(channel, "G") == 0) {
		for (int h=0; h < src_height; h++) {
			for (int w=0; w < src_width; w+=a) {
				b = *(srcp +w);
				g = *(srcp +w+1);
				r = *(srcp +w+2);
				m = (int) (sqrt(r*r + b*b) * (255.0/360.0)); // magenta
				*(dstp +w) = m;
				*(dstp +w+1) = g;
				*(dstp +w+2) = m;
			}															
			srcp += src_pitch;
			dstp += dst_pitch;
		}
	} else { // channel="B"
		for (int h=0; h < src_height; h++) {
			for (int w=0; w < src_width; w+=a) {
				b = *(srcp +w);
				g = *(srcp +w+1);
				r = *(srcp +w+2);
				y = (int) (sqrt(r*r + g*g) * (255.0/360.0)); // yellow
				*(dstp +w) = b;
				*(dstp +w+1) = y;
				*(dstp +w+2) = y;
			}															
			srcp += src_pitch;
			dstp += dst_pitch;
		}
	}
	return dst;
}

Sepia::Sepia(PClip _child, int _color, const char* _mode, IScriptEnvironment* env) :
	GenericVideoFilter(_child), color(_color), mode(_mode) {

	if (!vi.IsYUV()) {
		env->ThrowError("Sepia: your source is not YUV");
	}

	if ((lstrcmpi(mode, "linear") == 0) && (lstrcmpi(mode, "triangular") == 0)) {
		env->ThrowError("Sepia: mode must be \"linear\" or \"triangular\"");
	}

	int color_yuv = RGB2YUV(color);

	U = (color_yuv>>8) & 0xff;
	V = color_yuv & 0xff;
	}

Sepia::~Sepia() {
}


PVideoFrame __stdcall Sepia::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
		
	const unsigned char* srcp = src->GetReadPtr();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	
	PVideoFrame dst = env->NewVideoFrame(vi);				
	unsigned char* dstp = dst->GetWritePtr();
	const int dst_pitch = dst->GetPitch();

	// Sepia
	// r=162, g=138, b=101 (Y=137, U=108, V=141 => 0x896c8d)
	if (vi.IsYUY2()) {
		if (lstrcmpi(mode, "linear") == 0) {
			// equivalent with:
			// c = ImageSource("F:\CompilingPlugins\ColorLooks\before.jpg")
			// bw = GreyScale(c)
			// b = BlankClip(c, color=$a28a65)
			// Overlay(b, bw, mask=bw, opacity=1, mode="chroma")
			for (int h=0; h < src_height; h++) {
				for (int w=0; w < src_width; w+=4) {
					// (1-m)*bw_Y + m*b_c (m = bw_Y/255. b_c = (b_Y, b_U, b_V), b_Y=bw_Y=128)
					*(dstp +w) = *(srcp +w);
					*(dstp +w+1) = int( ( (255-srcp[w])*128 + srcp[w]*U ) ) >> 8;
					*(dstp +w+2) = *(srcp +w+2);
					*(dstp +w+3) = int( ( (255-srcp[w+2])*128 + srcp[w+2]*V ) ) >> 8;
				}															
				srcp += src_pitch;
				dstp += dst_pitch;
			}
		} else { // mode="triangular"
			for (int h=0; h < src_height; h++) {
				for (int w=0; w < src_width; w+=4) {
					// (1-m)*bw_Y + m*b_c (m = 0.5-abs((bw_Y-127.5)/255). b_c = (b_Y, b_U, b_V), b_Y=bw_Y=128)
					*(dstp +w) = *(srcp +w);
					*(dstp +w+1) = int( ( (127.5+fabs(srcp[w]-127.5))*128 + (127.5-fabs(srcp[w]-127.5))*U ) ) >> 8;
					*(dstp +w+2) = *(srcp +w+2);
					*(dstp +w+3) = int( ( (127.5+fabs(srcp[w+2]-127.5))*128 + (127.5-fabs(srcp[w+2]-127.5))*V ) ) >> 8;
				}															
				srcp += src_pitch;
				dstp += dst_pitch;
			}
		}
	} else { // YV12
		const int src_pitchUV = src->GetPitch(PLANAR_U);
		const int dst_pitchUV = dst->GetPitch(PLANAR_U);

		const unsigned char* srcpV = src->GetReadPtr(PLANAR_V);
		const unsigned char* srcpU = src->GetReadPtr(PLANAR_U);
		unsigned char* dstpV = dst->GetWritePtr(PLANAR_V);
		unsigned char* dstpU = dst->GetWritePtr(PLANAR_U);

		if (lstrcmpi(mode, "linear") == 0) {
			// equivalent with:
			// c = ImageSource("F:\CompilingPlugins\ColorLooks\before.jpg")
			// bw = GreyScale(c)
			// b = BlankClip(c, color=$a28a65)
			// Overlay(b, bw, mask=bw, opacity=1, mode="chroma")
			for (int h=0; h < src_height; h+=2) {
				for (int w=0; w < src_width; w+=2) {

					// (1-m)*bw_Y + m*b_c (m = bw_Y/255. b_c = (b_Y, b_U, b_V), b_Y=bw_Y=128)
					dstp[w] = srcp[w];
					dstp[w+1] = srcp[w+1];
					dstp[w+dst_pitch] = srcp[w+src_pitch];
					dstp[w+1+dst_pitch] = srcp[w+1+src_pitch];
					dstpV[w>>1] = int( ( (255-srcpV[w>>1])*128 + srcpV[w>>1]*U ) ) >> 8;
					dstpU[w>>1] = int( ( (255-srcpU[w>>1])*128 + srcpU[w>>1]*V ) ) >> 8;
				}
				srcp += src_pitch<<1;
				dstp += dst_pitch<<1;
				srcpV += src_pitchUV;
				srcpU += src_pitchUV;
				dstpV += dst_pitchUV;
				dstpU += dst_pitchUV;
			}
		} else { // mode="triangular"
			for (int h=0; h < src_height; h+=2) {
				for (int w=0; w < src_width; w+=2) {

					// (1-m)*bw_Y + m*b_c (m = 0.5-abs((bw_Y-127.5)/255). b_c = (b_Y, b_U, b_V), b_Y=bw_Y=128)
					dstp[w] = srcp[w];
					dstp[w+1] = srcp[w+1];
					dstp[w+dst_pitch] = srcp[w+src_pitch];
					dstp[w+1+dst_pitch] = srcp[w+1+src_pitch];
					dstpV[w>>1] = int( ( (127.5+fabs(srcpV[w>>1]-127.5))*128 + (127.5-fabs(srcpV[w>>1]-127.5))*U ) ) >> 8;
					dstpU[w>>1] = int( ( (127.5+fabs(srcpU[w>>1]-127.5))*128 + (127.5-fabs(srcpU[w>>1]-127.5))*V ) ) >> 8;
				}
				srcp += src_pitch<<1;
				dstp += dst_pitch<<1;
				srcpV += src_pitchUV;
				srcpU += src_pitchUV;
				dstpV += dst_pitchUV;
				dstpU += dst_pitchUV;
			}
		}
	}
	return dst;
}

Colorize::Colorize(PClip _child, int _color, bool _scale, IScriptEnvironment* env) :
	GenericVideoFilter(_child), color(_color), scale(_scale) {

	if (!vi.IsYUV()) {
		env->ThrowError("Colorize: your source is not YUV");
	}

	int color_yuv = RGB2YUV(color);

	U = (color_yuv>>8) & 0xff;
	V = color_yuv & 0xff;

	if (scale==1) {
		luma = (color_yuv>>16) & 0xff; // mimics VDub's Colorize
	} else {
		luma = 240;
	}
}

Colorize::~Colorize() {
}

PVideoFrame __stdcall Colorize::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
		
	const unsigned char* srcp = src->GetReadPtr();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	
	PVideoFrame dst = env->NewVideoFrame(vi);				
	unsigned char* dstp = dst->GetWritePtr();
	const int dst_pitch = dst->GetPitch();

	//------ Colorize -------
	// r=162, g=138, b=101 (Y=137, U=108, V=141)
	// rgb = 0xa28a65 (YUV = 0x896c8d)
	
	if (vi.IsYUY2()) {
		for (int h=0; h < src_height; h++) {
			for (int w=0; w < src_width; w+=4) {

				dstp[w] = int( srcp[w]*luma/240 );
				dstp[w+1] = U;
				dstp[w+2] = int( srcp[w+2]*luma/240 );
				dstp[w+3] = V;
			}															
			srcp += src_pitch;
			dstp += dst_pitch;
		}
	} else { // YV12
		const int src_pitchUV = src->GetPitch(PLANAR_U);
		const int dst_pitchUV = dst->GetPitch(PLANAR_U);

		const unsigned char* srcpV = src->GetReadPtr(PLANAR_V);
		const unsigned char* srcpU = src->GetReadPtr(PLANAR_U);
		unsigned char* dstpV = dst->GetWritePtr(PLANAR_V);
		unsigned char* dstpU = dst->GetWritePtr(PLANAR_U);

		for (int h=0; h < src_height; h+=2) {
			for (int w=0; w < src_width; w+=2) {

				dstp[w] = int( srcp[w]*luma/240 );
				dstp[w+1] = int( srcp[w+1]*luma/240 );
				dstp[w+dst_pitch] = int( srcp[w+src_pitch]*luma/240 );
				dstp[w+1+dst_pitch] = int( srcp[w+1+src_pitch]*luma/240 );

				dstpV[w>>1] = V;
				dstpU[w>>1] = U;
			}
			srcp += src_pitch<<1;
			dstp += dst_pitch<<1;
			srcpV += src_pitchUV;
			srcpU += src_pitchUV;
			dstpV += dst_pitchUV;
			dstpU += dst_pitchUV;
		}
	}
	return dst;
}

Posterize::Posterize(PClip _child, int _pbits, IScriptEnvironment* env) :
	GenericVideoFilter(_child), pbits(_pbits) {

	if (!(vi.IsRGB() || vi.IsYUV())) {
		env->ThrowError("Posterize: your source is not RGB or YUV");
	}

	// check: 0<=nbits<=8
	if ((pbits<0)||(pbits>8)) {
		env->ThrowError("Posterize: make sure that 0<=pbits<=8");
	}
}

Posterize::~Posterize() {
}

PVideoFrame __stdcall Posterize::GetFrame(int n, IScriptEnvironment* env) {

	PVideoFrame src = child->GetFrame(n, env);
		
	const unsigned char* srcp = src->GetReadPtr();
	const int src_pitch = src->GetPitch();
	const int src_width = src->GetRowSize();
	const int src_height = src->GetHeight();
	
	PVideoFrame dst = env->NewVideoFrame(vi);				
	unsigned char* dstp = dst->GetWritePtr();
	const int dst_pitch = dst->GetPitch();

	int np = 255 >> (8-pbits) << (8-pbits);

	int a;

	if (vi.IsRGB32()) {
		a=4;
	} else if (vi.IsRGB24()) {
		a=3;
	}

	if (vi.IsRGB()) {
		for (int h=0; h < src_height; h++) {
			for (int w=0; w < src_width; w+=a) {
				// should write it on the input clip !!! (holds for all filters i guess)
				dstp[w] = srcp[w]&np;
				dstp[w+1] = srcp[w+1]&np;
				dstp[w+2] = srcp[w+2]&np;
				//dstp[w+3] = srcp[w+3]&np;
			}															
			srcp += src_pitch;
			dstp += dst_pitch;
		}
	} else if (vi.IsYUY2()) {
		int npU = 127 >> (8-pbits) << (8-pbits);
		for (int h=0; h < src_height; h++) {
			for (int w=0; w < src_width; w+=4) {
				// should write it on the input clip !!! (holds for all filters i guess)
				dstp[w] = ( (srcp[w]-16)&np) + 16;
				dstp[w+2] = ( (srcp[w+2]-16)&np) + 16;
				if (srcp[w+1]>128) {
					dstp[w+1] = ( (srcp[w+1]-128) & npU ) + 128;
				} else {
					dstp[w+1] = - ( (128-srcp[w+1]) & npU ) + 128;
				}
				if (srcp[w+3]>128) {
					dstp[w+3] = ( (srcp[w+3]-128) & npU ) + 128;
				} else {
					dstp[w+3] = - ( (128-srcp[w+3]) & npU ) + 128;
				}
			}
			srcp += src_pitch;
			dstp += dst_pitch;
		}
	} else { // YV12
		const int src_pitchUV = src->GetPitch(PLANAR_U);
		const int dst_pitchUV = dst->GetPitch(PLANAR_U);

		const unsigned char* srcpV = src->GetReadPtr(PLANAR_V);
		const unsigned char* srcpU = src->GetReadPtr(PLANAR_U);
		unsigned char* dstpV = dst->GetWritePtr(PLANAR_V);
		unsigned char* dstpU = dst->GetWritePtr(PLANAR_U);

		int npU = 127 >> (8-pbits) << (8-pbits);

		for (int h=0; h < src_height; h+=2) {
			for (int w=0; w < src_width; w+=2) {
				// should write it on the input clip !!! (holds for all filters i guess)
				dstp[w] = ( (srcp[w]-16)&np) + 16;
				dstp[w+1] = ( (srcp[w+1]-16)&np) + 16;
				dstp[w+dst_pitch] = ( (srcp[w+src_pitch]-16)&np) + 16;
				dstp[w+1+dst_pitch] = ( (srcp[w+1+src_pitch]-16)&np) + 16;

				if (srcpV[w>>1]>128) {
					dstpV[w>>1] = ( (srcpV[w>>1]-128) & npU ) + 128;
				} else {
					dstpV[w>>1] = - ( (128-srcpV[w>>1]) & npU ) + 128;
				}
				if (srcpU[w>>1]>128) {
					dstpU[w>>1] = ( (srcpU[w>>1]-128) & npU ) + 128;
				} else {
					dstpU[w>>1] = - ( (128-srcpU[w>>1]) & npU ) + 128;
				}
			}
			srcp += src_pitch<<1;
			dstp += dst_pitch<<1;
			srcpV += src_pitchUV;
			srcpU += src_pitchUV;
			dstpV += dst_pitchUV;
			dstpU += dst_pitchUV;
		}
	}
	return dst;
}


AVSValue __cdecl Create_Technicolor(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new Technicolor(args[0].AsClip(), args[1].AsString("R"), env);
}

AVSValue __cdecl Create_Colorize(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new Colorize(args[0].AsClip(), args[1].AsInt(0xa28a65), args[2].AsBool(false), env);
}

AVSValue __cdecl Create_Sepia(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new Sepia(args[0].AsClip(), args[1].AsInt(0xa28a65), args[2].AsString("triangular"), env);
}

AVSValue __cdecl Create_Posterize(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new Posterize(args[0].AsClip(), args[1].AsInt(8), env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
  env->AddFunction("Technicolor", "c[channel]s", Create_Technicolor, 0);
  env->AddFunction("Colorize", "c[color]i[scale]b", Create_Colorize, 0);
  env->AddFunction("Sepia", "c[color]i[mode]s", Create_Sepia, 0);
  env->AddFunction("Posterize", "c[pbits]i", Create_Posterize, 0);
  return 0;
}
